<'
# Copyright (c) 2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar 14 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <stdio.h>

#include "<"$comp">/c/client.h"
#include "<"$comp">_portlib.h"
#include "<"$comp">_msglib.h"
#include "private.h"


<'foreach s [$component services] {'>
<'  if {[catch {$s task}]} continue'>

/* <"[--- ${comp}_[$s name]_action_cb -----------------------------------]"> */

void
<"$comp">_<"[$s name]">_action_cb(
  genom_client h, int rqstid, gnmix_<"[$s name]">::GoalHandle g)
{
  if (g.getCommState() == actionlib::CommState::DONE) {
    const boost::shared_ptr<const genom::srv::<"[$s name]">::output> rout =
      g.getResult();

    if (rout->report_)
      h->active_rqst[rqstid].error = GENOM_COMP_ERR_FLAG | rout->report_;
    h->active_rqst[rqstid].status = ROS_CLIENT_TERM;
    h->connections-><"[$s name]">_gh.remove(g);

<'  foreach p [$s parameter inout out] {'>
   genom::ids::pcopy(
     ((struct genom_<"$comp">_<"[$s name]">_output *)
      h->active_rqst[rqstid].output)-><"[$p name]">,
     rout-><"[$p name]">);
<'  }'>
  }
}

/* <"[--- ${comp}_[$s name]_action_feedback_cb --------------------------]"> */

void
<"$comp">_<"[$s name]">_action_feedback_cb(
  genom_client h, int rqstid, gnmix_<"[$s name]">::GoalHandle g,
  const boost::shared_ptr<const uint32_t> aid)
{
  h->active_rqst[rqstid].aid = *aid;
  h->active_rqst[rqstid].status = ROS_CLIENT_WAIT_TERM;
}
<'}'>

<'foreach s [$component services] {'>

/* <"[--- ${comp}_[$s name]_rqst ----------------------------------------]"> */

int
genom_<"$comp">_client_<"[$s name]">_rqst(
  genom_client h,
  const struct genom_<"$comp">_<"[$s name]">_input *in,
  <"$comp">_<"[$s name]">_cb sentcb, <"$comp">_<"[$s name]">_cb donecb,
  void *cb_data,
  int *rqstid)
{
<'  set svc [catch {$s task}]'>
  genom::srv::<"[$s name]">::input rin;
  int s;
  assert(h);

<'  foreach p [$s parameter in inout] {'>
  genom::ids::pcopy(rin.<"[$p name]">, in-><"[$p name]">);
<'  }'>

  h->error = ros_newrqst(
    h, sizeof(struct genom_<"$comp">_<"[$s name]">_output), rqstid);
  if (h->error) return h->error;

  genom_<"$comp">_client_<"[$s name]">_init_output(
    (struct genom_<"$comp">_<"[$s name]">_output *)
    h->active_rqst[*rqstid].output);
  h->active_rqst[*rqstid].info =
    genom_<"$comp">_client_<"[$s name]">_info;
  h->active_rqst[*rqstid].sentcb = (genom_request_cb)sentcb;
  h->active_rqst[*rqstid].donecb = (genom_request_cb)donecb;
  h->active_rqst[*rqstid].cb_data = cb_data;
  h->active_rqst[*rqstid].pstatus = ROS_CLIENT_WAIT_ACK;

<'  if {$svc} {'>
  h->active_rqst[*rqstid].status = ROS_CLIENT_WAIT_TERM;
  genom::srv::<"[$s name]">::output rout;
  if (!h->connections-><"[$s name]">.call(rin, rout))
    h->active_rqst[*rqstid].error = GENOM_SYS_ERR_FLAG | ENXIO;
  else {
<'    foreach p [$s parameter inout out] {'>
    genom::ids::pcopy(
      ((struct genom_<"$comp">_<"[$s name]">_output *)
       h->active_rqst[*rqstid].output)-><"[$p name]">,
      rout.<"[$p name]">);
<'    }'>
    if (rout.report_)
      h->active_rqst[*rqstid].error = GENOM_COMP_ERR_FLAG | rout.report_;
  }
  h->active_rqst[*rqstid].status = ROS_CLIENT_TERM;

<'  } else {'>

  h->active_rqst[*rqstid].status = ROS_CLIENT_WAIT_ACK;
  h->connections-><"[$s name]">_gh.push_front(
    h->connections-><"[$s name]">->sendGoal(
      rin, boost::bind(&<"$comp">_<"[$s name]">_action_cb, h, *rqstid, _1),
      boost::bind(&<"$comp">_<"[$s name]">_action_feedback_cb,
                  h, *rqstid, _1, _2)));
<'  }'>

<'  if {!$svc} {'>
  if (!sentcb) {
    s = ros_waitack(h, *rqstid);
    if (s) return s;
  }
<'  }'>

  if (!donecb) {
    s = ros_waitterm(h, *rqstid);
    if (s) return s;
  }

  return 0;
}

<'}'>

<'foreach p [$component ports out] {'>

/* <"[--- ${comp}_[$p name]_port_cb -------------------------------------]"> */

void
<"$comp">_<"[$p name]">_port_cb(
  int &done,
  <"[[$p datatype] argument reference out]">,
  const boost::shared_ptr<genom::port::<"[$p name]"> const> &data)
{
  genom::ids::pcopy(<"[[$p datatype] dereference out]">, data->v);
  done = 1;
}
<'}'>

<'foreach p [$component ports out simple] {'>

/* <"[--- ${comp}_[$p name]_port ----------------------------------------]"> */

int
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
  <"[[$p datatype] argument reference data]">)
{
  ros::Subscriber s;
  int delay, done = 0;

  s = h->node->subscribe< genom::port::<"[$p name]"> >(
    "/" + std::string(h->instance) + "/<"[$p name]">", 1,
    boost::bind(&<"$comp">_<"[$p name]">_port_cb, boost::ref(done), data, _1));
  if (!s) return h->error = GENOM_SYS_ERR_FLAG | ENOENT;

  delay = 0;
  while(!done) {
    struct timespec req = {0, 1000000/*1ms*/}, rem;
    delay ++;
    if (delay > 3000/*3s*/) return h->error = GENOM_SYS_ERR_FLAG | ENOENT;
    while (nanosleep(&req, &rem) == EINTR) req = rem;
  }

  return 0;
}

<'}'>
<'foreach p [$component ports out multiple] {'>

/* <"[--- ${comp}_[$p name]_port ----------------------------------------]"> */

int
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
  const char *path,
  <"[[$p datatype] argument reference data]">)
{
  ros::Subscriber s;
  int delay, done = 0;

  s = h->node->subscribe< genom::port::<"[$p name]"> >(
    "/" + std::string(h->instance) + "/<"[$p name]">/" + path, 1,
    boost::bind(&<"$comp">_<"[$p name]">_port_cb, boost::ref(done), data, _1));
  if (!s) return h->error = GENOM_SYS_ERR_FLAG | ENOENT;

  delay = 0;
  while(!done) {
    struct timespec req = {0, 1000000/*1ms*/}, rem;
    delay ++;
    if (delay > 3000/*3s*/) return h->error = GENOM_SYS_ERR_FLAG | ENOENT;
    while (nanosleep(&req, &rem) == EINTR) req = rem;
  }

  return 0;
}

<'}'>
