<'
# Copyright (c) 2011-2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Tue Dec 13 2011
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include <cstring>

#include "ros/ros.h"
#include "ros/callback_queue.h"
#include "actionlib/server/action_server.h"

#include "<"$comp">_internals.h"
#include "<"$comp">_typecopy.h"


/* --- local data ---------------------------------------------------------- */

<"$comp">CntrlStrId cids;


/* <"[--- ${comp}_cntrl_task ----------------------------------------------- ]"> */

/** Control task of component <"$comp">
 *
 * This task waits for incoming messages or internal events. Upon reception
 * of a request, the corresponding callback function is invoked. Internal
 * events are handled and final replies are sent to the client.
 */

int
<"$comp">_cntrl_task(const char *instance, int signalfd)
{
  ros::CallbackQueue *queue = ros::getGlobalCallbackQueue();
  pthread_mutexattr_t attr;
  int s;

  cids.shutdown = 0;
  cids.node = new ros::NodeHandle(instance);

  s = pthread_mutexattr_init(&attr);
  s += pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
  s += pthread_mutex_init(&cids.task_lock, &attr);
  if (s) err(2, "initializing task lock");

  cids.ticks = 0;
  memset(cids.run_map, 0, sizeof(cids.run_map));
<'if {![catch {$component ids}]} {'>
  genom::ids::pinit(cids.ids);
<'}'>

  /* advertise data outports */
  genom_log_info("advertising ports");
<'foreach p [$component ports out simple] {'>
  cids.ports.<"[$p name]">.init(cids.node);
<'}'>
  s = genom_state_<"$comp">_init();
  if (s) errx(2, "error while initializing state port");

  /* create tasks */
<'foreach t [$component tasks] {'>
  cids.tasks.<"[$t name]">.spawn();
<'}'>

  /* advertise services */
  genom_log_info("advertising services");
<'foreach s [$component services] {'>
<'  if {[catch {$s task} t]} {'>
  cids.services.<"[$s name]"> =
    cids.node->advertiseService("<"[$s name]">",
                                <"$comp">_<"[$s name]">_rqstcb);
<'  } else {'>
  cids.services.<"[$s name]"> =
    new actionlib::ActionServer< genom::action::<"[$s name]"> >(
      *cids.node, "<"[$s name]">",
      <"$comp">_<"[$s name]">_rqstcb, <"$comp">_<"[$s name]">_intercb,
      false);
  cids.services.<"[$s name]">->start();
<'  }'>
<'}'>

<'if {![catch {$component clock-rate} rate]} {'>
  /* create timer */
  cids.clock = cids.node->createTimer(ros::Duration(<"[$rate value]">),
                                      <"$comp">_timercb,
                                      false, true);
<'}'>

  /* signal parent in daemon mode */
  if (signalfd >= 0) {
    ssize_t dummy = write(signalfd, "x", 1);
    if (dummy) /* silence gcc warnings */;
    close(signalfd);
  }

  /* go */
  genom_log_info("control task initialized and running");
  while(ros::ok())
    queue->callAvailable(ros::WallDuration(1.));

  /* cleanup */
<'if {![catch {$component clock-rate} rate]} {'>
  cids.clock.stop();
<'}'>
  pthread_mutex_lock(&cids.task_lock);
  cids.shutdown = 1;
<'foreach task [$component tasks] {'>
  pthread_cond_signal(&cids.tasks.<"[$task name]">.signal);
<'}'>
  pthread_mutex_unlock(&cids.task_lock);
<'foreach task [$component tasks] {'>
  pthread_join(cids.tasks.<"[$task name]">.id, NULL);
<'}'>

  genom_log_info("shutting down services");
<'foreach s [$component services] {'>
<'  if {[catch {$s task} t]} {'>
  cids.services.<"[$s name]">.shutdown();
<'  } else {'>
  delete cids.services.<"[$s name]">;
<'  }'>
<'}'>

  genom_log_info("shutting down ports");
<'foreach p [$component ports] {'>
  cids.ports.<"[$p name]">.fini();
<'}'>

<'if {![catch {$component ids}]} {'>
  genom::ids::pfini(cids.ids);
<'}'>
  cids.node->shutdown();
  delete cids.node;
  return 0;
}


/* <"[--- ${comp}_cntrl_task_signal -------------------------------------]"> */

/** \brief Signal the control task that an activity is terminated
 */

static const struct genom_activity_cbd_ {
  void operator()(genom_activity *) {}
} genom_activity_cbd = {};

void
<"$comp">_cntrl_task_signal(genom_activity *a)
{
  ros::CallbackInterfacePtr cb(a, genom_activity_cbd);
  ros::getGlobalCallbackQueue()->addCallback(cb);
}


/* <"[--- ${comp}_exec_task_signal --------------------------------------]"> */

/** \brief Signal executions tasks
 */

void
<"$comp">_exec_task_signal(void)
{
<'foreach t [$component tasks] {'>
  if (!cids.tasks.<"[$t name]">.runnable) {
    cids.tasks.<"[$t name]">.runnable = true;
<'  if {[catch {$t period}]} {'>
    cids.tasks.<"[$t name]">.wakeup = true;
    pthread_cond_signal(&cids.tasks.<"[$t name]">.signal);
<'  }'>
  }
<'}'>
}


/* --- Timer callback ------------------------------------------------------ */

void
<"$comp">_timercb(const ros::TimerEvent &e)
{
  pthread_mutex_lock(&cids.task_lock);
  cids.ticks++;
<'foreach task [$component tasks] {'>
<'  if {![catch {$task period} period]} {'>
<'
    set rate [[$component clock-rate] value]
    set ticks [expr {int([$period value]/$rate)}]
    if {![catch {$task delay} delay]} {
      set d [expr {int([$delay value]/$rate)}]
    } else {
      set d 0
    }
'>
  if (cids.tasks.<"[$task name]">.runnable &&
      cids.ticks % <"$ticks"> == <"$d">) {
    cids.tasks.<"[$task name]">.wakeup = true;
    pthread_cond_signal(&cids.tasks.<"[$task name]">.signal);
  }
<'  }'>
<'}'>
  pthread_mutex_unlock(&cids.task_lock);
}


<'foreach s [$component services] {'>


<'  if {[catch {$s task} task]} {'>
/* <"[--- Control service [$s name] ----------------------------------------]"> */

bool
<"$comp">_<"[$s name]">_rqstcb(
  genom::srv::<"[$s name]">::input &in, genom::srv::<"[$s name]">::output &out)
{
  pthread_mutex_lock(&cids.task_lock);

<'    if {[llength [$s interrupt]]} {'>
  /* interrupt incompatible activities */
  for(size_t i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
    genom_activity *c = cids.activities.activities[i];
    if (!c) continue;
    switch(c->rqstid) {
<'      foreach c [$s interrupt] {'>
      case <"$COMP">_<"[$c name]">_RQSTID:
<'      }'>
        switch(c->status) {
          case ACT_ETHER:
          case ACT_INTER:
          case ACT_STOP:
            break;

          case ACT_WAIT:
            c->reject(<"[[[$component event] member stop] declarator]">);
            cids.activities.free(c);
            delete c;
            break;

          case ACT_RUN:
            c->stop();
            <"$comp">_exec_task_signal();
            break;
        }
        break;

      default: break;
    }
  }
<'    }'>

  out.report_ = <"$comp">_<"[$s name]">_controlcb(&cids, in, out);
  <"$comp">_exec_task_signal(); /* wake up sleeping activities */
  if (out.report_ != <"[[[$component event] member ok] declarator]">) {
    pthread_mutex_unlock(&cids.task_lock);
    return true;
  }

  cids.run_map[<"$COMP">_<"[$s name]">_RQSTID] = true;
  pthread_mutex_unlock(&cids.task_lock);
  return true;
}
<'  } else {'>
/* <"[--- Execution service [$s name] --------------------------------------]"> */

template<> int
genom_activity_data< genom::action::<"[$s name]"> >::interrupt()
{
  int wait = 0;
<'    if {[llength [$s interrupt]]} {'>
  genom_activity *c;
  size_t i;

  /* interrupt incompatible activities */
  for(i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
    c = cids.activities.activities[i];
    if (!c || c->id == id) continue;
    switch(c->rqstid) {
<'      foreach c [$s interrupt] {'>
      case <"$COMP">_<"[$c name]">_RQSTID:
<'      }'>
        switch(c->status) {
          case ACT_ETHER:
          case ACT_INTER:
            break;

          case ACT_STOP:
            wait = 1;
            break;

          case ACT_WAIT:
            c->reject(<"[[[$component event] member stop] declarator]">);
            cids.activities.free(c);
            delete c;
            break;

          case ACT_RUN:
            c->stop();
            <"$comp">_exec_task_signal();
            wait = 1;
            break;
        }
        break;

      default: break;
    }
  }
<'    }'>

  return wait;
}

template<> <"[[$component event] declarator]">
genom_activity_data< genom::action::<"[$s name]"> >::validate()
{
  <"[[$component event] declarator s]">;

  s = <"$comp">_<"[$s name]">_controlcb(&cids, in, out);
  if (s != <"[[[$component event] member ok] declarator]">)
    return s;

  cids.run_map[<"$COMP">_<"[$s name]">_RQSTID] = true;
  return s;
}

void
<"$comp">_<"[$s name]">_rqstcb(
  actionlib::ServerGoalHandle< genom::action::<"[$s name]"> > &rqst)
{
  <"[[$component event] declarator s]">;
  genom_activity_data< genom::action::<"[$s name]"> > *a;

  a = new genom_activity_data< genom::action::<"[$s name]"> >(
    <"$COMP">_<"[$s name]">_RQSTID, cids.tasks.<"[$task name]">.id,
    rqst);

  pthread_mutex_lock(&cids.task_lock);
  s = cids.activities.alloc(a);
  if (s != <"[[[$component event] member ok] declarator]">) {
    a->reject(s);
    pthread_mutex_unlock(&cids.task_lock);
    delete a;
    return;
  }

  if (a->interrupt()) {
    /* wait for other activities */
    pthread_mutex_unlock(&cids.task_lock);
    return;
  }

  s = a->accept();
  if (s != <"[[[$component event] member ok] declarator]">) {
    a->reject(s);
    cids.activities.free(a);
    pthread_mutex_unlock(&cids.task_lock);
    delete a;
    return;
  }

  <"$comp">_exec_task_signal();
  pthread_mutex_unlock(&cids.task_lock);
}

void
<"$comp">_<"[$s name]">_intercb(
  actionlib::ServerGoalHandle< genom::action::<"[$s name]"> > &rqst)
{
  actionlib_msgs::GoalID gid;
  genom_activity *a;

  pthread_mutex_lock(&cids.task_lock);
  gid = rqst.getGoalID();
  a = cids.activities.bygid(gid);
  if (a == NULL) {
    pthread_mutex_unlock(&cids.task_lock);
    genom_log_warn(
      "cancel request for non-existent activity %s", gid.id.c_str());
    return;
  }

  a->stop();
  <"$comp">_exec_task_signal();
  pthread_mutex_unlock(&cids.task_lock);
}

<'  }'>
<'}'>


/* --- Activity callback --------------------------------------------------- */

void
<"$comp">_cntrl_task_activitycb(genom_activity *a)
{
  <"[[$component event] declarator s]">;
  genom_activity *c;
  size_t i;

  pthread_mutex_lock(&cids.task_lock);
  a->report();
  cids.activities.free(a);
  <"$comp">_exec_task_signal();

  /* check for pending activities */
  for(i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
    c = cids.activities.activities[i];
    if (!c || c->status != ACT_WAIT) continue;
    if (c->interrupt()) continue;

    s = c->accept();
    if (s != <"[[[$component event] member ok] declarator]">) {
      c->reject(s);
      cids.activities.free(c);
      delete c;
    }
  }

  pthread_mutex_unlock(&cids.task_lock);
  delete a;
}
