<'
# Copyright (c) 2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Sat Jan 14 2012
#

if {[llength $argv] != 2} { error "expected arguments: component task" }
lassign $argv component task

# compute handy shortcuts
set comp [$component name]
set tname [$task name]

lang c
'>
#include "autoconf/acheader.h"

#include "<"$comp">_internals.h"


/* --- local data ---------------------------------------------------------- */

static void *	<"$comp">_<"$tname">_task(void *data);
<'if {[llength [$task fsm]]} {'>
static int	<"$comp">_<"$tname">_invoke(
			<"[[$component event] declarator &state]">);
<'}'>


/* <"[--- genom_task_${tname}::spawn ------------------------------------]"> */

void
genom_task_<"$tname">::spawn()
{
  int s;

  s = pthread_cond_init(&signal, NULL);
  if (s) err(2, "initializing task condition variable");
  id = 0;

  pthread_mutex_lock(&cids.task_lock);
  runnable = wakeup = false;
  status = <"[[[$component event] member sleep] declarator]">;

  s = pthread_create(&id, NULL, <"$comp">_<"$tname">_task, NULL);
  if (s) err(2, "spawning task <"[$t name]">");
  do {
    s = pthread_cond_wait(&signal, &cids.task_lock);
    if (s) err(2, "spawning task <"[$t name]">");
  } while(status == <"[[[$component event] member sleep] declarator]">);
  pthread_mutex_unlock(&cids.task_lock);
}


/* <"[--- ${comp}_${tname}_task -----------------------------------------]"> */

/** <"$tname"> task of component <"$comp">
 *
 */

static void *
<"$comp">_<"$tname">_task(void *dummy)
{
  genom_task_<"$tname"> &task = cids.tasks.<"$tname">;
  genom_activity *a;
  size_t i;
  int s;

  pthread_mutex_lock(&cids.task_lock);
  task.status = <"[[[$component event] member ok] declarator]">;
  task.pact = <"[[[$component event] member start] declarator]">;

<'if {[llength [$task fsm]]} {'>
  /* permanent activity start codel */
  s = <"$comp">_<"$tname">_invoke(task.pact);
  if (s) task.status = task.pact;
<'}'>

  genom_log_info("<"$tname"> task initialized and running");
  pthread_cond_signal(&task.signal);
  pthread_mutex_unlock(&cids.task_lock);

  if (task.status != <"[[[$component event] member ok] declarator]">) {
    genom_log_warn("shutting down <"$tname"> task");
    return NULL;
  }

  /* loop */
  pthread_mutex_lock(&cids.task_lock);
  do {
    while(!task.wakeup && !cids.shutdown) {
      s = pthread_cond_wait(&task.signal, &cids.task_lock);
      if (s) err(2, "timer task <"[$t name]">");
    }
    if (cids.shutdown) break;
    task.runnable = task.wakeup = false;
    genom_log_debug("<"$tname"> tick %d", cids.ticks);

<'if {[llength [$task fsm]]} {'>
    /* handle permanent activity */
    if (task.pact != <"[[[$component event] member ether] declarator]">) {
      s = <"$comp">_<"$tname">_invoke(task.pact);
      if (s) {
        task.status = task.pact; break;
      }
    }
<'}'>

    /* handle current activities for this task */
    for(i=0; i<genom_activities::MAX_ACTIVITIES; i++) {
      a = cids.activities.activities[i];
      if (!a || a->task != task.id) continue;
      if (a->status != ACT_RUN && a->status != ACT_STOP) continue;

      genom_log_debug("<"$tname"> activity %d", a->id);
      a->invoke();
    }
    pthread_mutex_unlock(&cids.task_lock);
    sched_yield();
    pthread_mutex_lock(&cids.task_lock);
  } while(1);

<'if {[llength [$task fsm stop]]} {'>
  /* stop permanent activity */
  if (task.pact != <"[[[$component event] member ether] declarator]">) {
    task.pact = <"[[[$component event] member stop] declarator]">;
    do {
      s = <"$comp">_<"$tname">_invoke(task.pact);
      if (s) {
        task.status = task.pact; break;
      }
    } while(task.pact != <"[[[$component event] member ether] declarator]">);
  }
<'}'>
  pthread_mutex_unlock(&cids.task_lock);

  genom_log_info("shutting down <"$tname"> task");
  return NULL;
}


/* === task codels ========================================================= */
<'
foreach e [$task fsm] {
  set codel [$task fsm $e]
'>

/* state <"$e">  */
static <"[[$component event] declarator]">
<"$comp">_tcodel_<"$tname">_<"$e">()
{
  <"[[$component event] declarator]"> s;
<'  # copy parameters for async codels
  if {[$codel kind] == "async"} {
    foreach p [$codel parameters ids] {'>
  <"[[$p type] declarator [$p name]_]">;
<'    }'>

<'    foreach p [$codel parameters ids] {'>
  genom::ids::pinit(<"[$p name]_">);
  genom::ids::pcopy(<"[$p name]_">, cids.ids<"[$p member]">);
<'    }'>
<'  }'>

<'  # build parameter list
  set plist [list]
  foreach p [$codel parameters] {
    switch -glob -- "[$codel kind]/[$p src]" {
      */local	{ error "task codel with service parameter" }
      */remote	{ set m "cids.remotes.[[$p remote] name].handle" }
      */port	{ set m "cids.ports.[[$p port] name].handle" }
      async/ids	{ set m [$p name]_ }
      sync/ids	{ set m "cids.ids[$p member]" }
      default	{ error "reached unreachable code" }
    }

    switch  -glob -- "[$p dir]/[$p src]" {
      in/* - */port {
        lappend plist [[$p type] pass value $m]
      }
      default {
        lappend plist [[$p type] pass reference $m]
      }
    }
  }
'>
<'if {[$codel kind] == "async"} {'>
  pthread_mutex_unlock(&cids.task_lock);
<'}'>
  s = <"[$codel invoke $plist]">;
<'if {[$codel kind] == "async"} {'>
  pthread_mutex_lock(&cids.task_lock);

<'  foreach p [$codel parameters ids] {'>
<'    if {[$p dir] != "in"} {'>
  genom::ids::pcopy(cids.ids<"[$p member]">, <"[$p name]_">);
<'    }'>
  genom::ids::pfini(<"[$p name]_">);
<'  }'>
<'}'>
  return s;
}
<'}'>


<'if {[llength [$task fsm]]} {'>
/* invoke codels according to state */
static int
<"$comp">_<"$tname">_invoke(<"[[$component event] declarator &state]">)
{
  <"[[$component event] declarator s]">;

  switch(state) {
    case <"[[[$component event] member ether] declarator]">:
<'  if {[lsearch [$task fsm] sleep] < 0} {'>
    case <"[[[$component event] member sleep] declarator]">:
<'  }'>
      break;

<'  foreach e [$task fsm] {'>
    case <"[[[$component event] member $e] declarator]">:
      s = <"$comp">_tcodel_<"$tname">_<"$e">();
      switch(s) {
<'    foreach y [[$task fsm $e] yield] {'>
        case <"[$y declarator]">:
<'    }'>
          state = s;
<'    foreach p [[$task fsm $e] parameters port] {'>
<'      if {"handle" ni [[$p port] kind] && "out" in [[$p port] kind]} {'>
          cids.ports.<"[$p name]">.publish();
<'      }'>
<'    }'>
          if (s != <"[[[$component event] member sleep] declarator]">)
            <"$comp">_exec_task_signal();
          break;

<'    if {[llength [$task throw]]} {'>
<'      foreach t [$task throw] {'>
        case <"[$t declarator]">:
<'      }'>
          state = s;
          return -1;
<'    }'>

        default:
          state = <"[[[$component event] member bad_transition] declarator]">;
          return -1;
      }
      break;

<'  }'>
    default: assert(!"Bad state id in task <"$tname">, permanent activity");
  }
  return 0;
}
<'}'>


/* === services codels ===================================================== */

<'foreach s [$task services] {'>
/* <"[--- Service [$s name] ---------------------------------------------]"> */

<'  foreach e [$s fsm] { set codel [$s fsm $e]'>

/* state <"$e"> */
static inline <"[[$component event] declarator]">
<"$comp">_codel_<"[$s name]">_<"$e">(
  genom_activity_data< genom::action::<"[$s name]"> > *a)
{
  <"[[$component event] declarator]"> s;

<'  # copy parameters for async codels
  if {[$codel kind] == "async"} {
    foreach p [$codel parameters ids] {'>
  <"[[$p type] declarator [$p name]_]">;
<'    }'>

<'    foreach p [$codel parameters ids] {'>
  genom::ids::pinit(<"[$p name]_">);
  genom::ids::pcopy(<"[$p name]_">, cids.ids<"[$p member]">);
<'    }'>
<'  } '>

<'    # build parameter list
    set plist [list]
    foreach p [$codel parameters] {
      switch -glob -- "[$codel kind]/[$p src]/[$p dir]" {
        */local/in	{ set m "a->in.[$p name]" }
        */local/*	{ set m "a->out.[$p name]" }
        */remote/*	{ set m "cids.remotes.[[$p remote] name].handle" }
        */port/*	{ set m "cids.ports.[[$p port] name].handle" }
        async/ids/*	{ set m [$p name]_ }
        sync/ids/*	{ set m "cids.ids[$p member]" }
        default		{ error "reached unreachable code" }
      }

      switch  -glob -- "[$p dir]/[$p src]" {
	in/* - */port {
          lappend plist [[$p type] pass value $m]
        }
	default {
	  lappend plist [[$p type] pass reference $m]
	}
      }
    }'>
<'if {[$codel kind] == "async"} {'>
  pthread_mutex_unlock(&cids.task_lock);
<'}'>
  s = <"[$codel invoke $plist]">;
<'if {[$codel kind] == "async"} {'>
  pthread_mutex_lock(&cids.task_lock);

<'  foreach p [$codel parameters ids] {'>
<'    if {[$p dir] != "in"} {'>
  genom::ids::pcopy(cids.ids<"[$p member]">, <"[$p name]_">);
<'    }'>
  genom::ids::pfini(<"[$p name]_">);
<'  }'>
<'}'>
  return s;
}

<'  }'>
/* invoke codels according to <"[$s name]"> state */
template<> int
genom_activity_data< genom::action::<"[$s name]"> >::invoke()
{
  <"[[$component event] declarator s]">;

  if (interrupted) {
    interrupted = 0;
    status = ACT_STOP;
    state = <"[[[$component event] member stop] declarator]">;
  }

  switch(state) {
<'  if {[lsearch [$s fsm] sleep] < 0} {'>
    case <"[[[$component event] member sleep] declarator]">:
<'  }'>

<'  if {[lsearch [$s fsm] stop] < 0} {'>
    case <"[[[$component event] member stop] declarator]">:
      state = <"[[[$component event] member ether] declarator]">;
      done();
      return -1;
<'  }'>

<'  foreach e [$s fsm] {'>
    case <"[[[$component event] member $e] declarator]">:
      s = <"$comp">_codel_<"[$s name]">_<"$e">(this);
      switch(s) {
<'    foreach y [[$s fsm $e] yield] {'>
        case <"[$y declarator]">:
<'    }'>
          state = s;
<'    foreach p [[$s fsm $e] parameters port] {'>
<'      if {"handle" ni [[$p port] kind] && "out" in [[$p port] kind]} {'>
          cids.ports.<"[$p name]">.publish();
<'      }'>
<'    }'>
          if (s == <"[[[$component event] member ether] declarator]">) {
            done();
          } else if (s != <"[[[$component event] member sleep] declarator]">) {
            <"$comp">_exec_task_signal();
          }
          break;

<'    foreach t [$s throw] {'>
        case <"[$t declarator]">:
          state = s;
          done();
          return -1;
<'    }'>

        default:
          state = <"[[[$component event] member bad_transition] declarator]">;
          done();
          return -1;
      }
      break;

<'  }'>
    default: assert(!"Bad state id in task <"$tname">, service <"[$s name]">");
  }
  return 0;
}

<'}'>
